/*----------------------------------------------------------------------
  File    : sodim.c
  Contents: subset size occurrence distribution based item set mining
  Author  : Christian Borgelt
  History : 2010.10.25 file created from file jim.c
            2010.11.24 adapted to modified error reporting (tract)
            2010.12.11 adapted to a generic error reporting function
            2010.12.20 adapted to function tbg_icnts() (filter problem)
            2011.03.20 optional integer transaction weights added
            2011.05.06 optional output of transaction identifiers added
            2011.05.10 filtering for closed and maximal item sets added
            2011.05.16 similarity measures generalized to ext. support
            2011.05.17 binary logarithm of support quotient added
            2011.07.08 adapted to modified function tbg_recode()
            2011.08.11 bug in function ispseudo() fixed (negative sfrac)
            2011.08.28 output of item set counters per size added
            2012.07.05 bug introduced by exchaning n and k fixed
            2013.03.29 adapted to type changes in module tract
            2013.10.15 checks of return code of isr_report() added
            2013.10.18 optional pattern spectrum collection added
            2013.11.12 item selection file changed to option -R#
            2014.08.26 adapted to modified item set reporter interface
            2014.08.28 functions sod_data() and sod_report() added
            2014.10.24 changed from LGPL license to MIT license
            2016.03.16 bug in item set reporter setup fixed (smin)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#ifndef ISR_PATSPEC
#define ISR_PATSPEC
#endif
#ifndef ISR_CLOMAX
#define ISR_CLOMAX
#endif
#ifdef SODIM_MAIN
#ifndef PSP_REPORT
#define PSP_REPORT
#endif
#ifndef TA_READ
#define TA_READ
#endif
#endif
#include "sodim.h"
#include "params.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "sodim"
#define DESCRIPTION "subset size occurrence distribution " \
                    "based item set mining"
#define VERSION     "version 4.19 (2016.10.15)        " \
                    "(c) 2010-2016   Christian Borgelt"

/* --- error codes --- */
/* error codes   0 to  -4 defined in tract.h */
#define E_STDIN      (-5)       /* double assignment of stdin */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* too few/many arguments */
#define E_TARGET     (-9)       /* invalid target type */
#define E_SIZE      (-10)       /* invalid item set size */
#define E_SUPPORT   (-11)       /* invalid minimum support */
#define E_SIMILTY   (-12)       /* invalid similarity measure */
#define E_MEASURE   (-13)       /* invalid evaluation measure */
/* error codes -15 to -25 defined in tract.h */

#ifndef QUIET                   /* if not quiet version, */
#define MSG         fprintf     /* print messages */
#define XMSG        if (mode & SOD_VERBOSE) fprintf
#define CLOCK(t)    ((t) = clock())
#else                           /* if quiet version, */
#define MSG(...)    ((void)0)   /* suppress messages */
#define XMSG(...)   ((void)0)
#define CLOCK(t)    ((void)0)
#endif

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef double SIMFN (double s, double r, double n, double *p);

typedef struct {                /* --- recursion data --- */
  SUPP     smin;                /* minimum support of an item set */
  TID      **lists;             /* transaction identifier lists */
  SUPP     *muls;               /* multiplicities of the transactions */
  ITEM     *cnts;               /* item counters, one per transaction */
  double   *dist;               /* subset size occurrence distrib. */
  double   *xsps;               /* extended support parameters */
  ITEM     xspcnt;              /* number of ext. subset parameters */
  SIMFN    *simfn;              /* similarity measure function */
  double   simps[8];            /* similarity measure parameters */
  double   sim;                 /* minimum similarity measure value */
  double   sfrac;               /* minimum cond. support fraction */
  ITEMBASE *base;               /* underlying item base */
  ISREPORT *report;             /* item set reporter */
} RECDATA;                      /* (recursion data) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#if !defined QUIET && defined SODIM_MAIN
/* --- error messages --- */
static const char *errmsgs[] = {
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_TARGET   -9 */  "invalid target type '%c'",
  /* E_SIZE    -10 */  "invalid item set size %"ITEM_FMT,
  /* E_SUPPORT -11 */  "invalid minimum support %g",
  /* E_SIMILTY -12 */  "invalid similarity measure '%c'",
  /* E_MEASURE -13 */  "invalid evaluation measure '%c'",
  /*           -14 */  NULL,
  /* E_NOITEMS -15 */  "no (frequent) items found",
  /*           -16 */  "unknown error"
};
#endif

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
#ifdef SODIM_MAIN
#ifndef QUIET
static CCHAR    *prgname;       /* program name for error messages */
#endif
static TABREAD  *tread  = NULL; /* table/transaction reader */
static ITEMBASE *ibase  = NULL; /* item base */
static TABAG    *tabag  = NULL; /* transaction bag/multiset */
static ISREPORT *report = NULL; /* item set reporter */
static TABWRITE *twrite = NULL; /* table writer for pattern spectrum */
static double   *border = NULL; /* support border for filtering */
#endif

/*----------------------------------------------------------------------
  Similarity Measure Functions
----------------------------------------------------------------------*/

static double constant        (double s, double r, double n, double *p)
{ return 0; }

static double russel_rao      (double s, double r, double n, double *p)
{ return s/n; }
/* S_R = s/n */

static double kulczynski      (double s, double r, double n, double *p)
{ double q = r-s; return (q > 0) ? s/q : INFINITY; }
/* S_K = s/q = s/(r-s) */

static double jaccard         (double s, double r, double n, double *p)
{ return s/r; }
/* S_J = s/r */

static double dice            (double s, double r, double n, double *p)
{ return (s+s)/(r+s); }
/* S_D = 2s/(r+s) */

static double sokal_sneath_1  (double s, double r, double n, double *p)
{ double d = r+r-s; return (d > 0) ? s/d : INFINITY; }
/* S_S = s/(r+q) = s/(2r-s) */

static double sokal_michener  (double s, double r, double n, double *p)
{ return (s+(n-r))/n; }
/* S_M = (s+z)/n */

static double faith           (double s, double r, double n, double *p)
{ return (s+0.5*(n-r))/n; }
/* S_F = (s+z/2)/n */

static double rogers_tanimoto (double s, double r, double n, double *p)
{ double q = r-s; return (n-q)/(n+q); }
/* S_T = (s+z)/(n+q) = (n-q)/(n+q) */

static double sokal_sneath_2  (double s, double r, double n, double *p)
{ double sz = s+(n-r); return (sz+sz)/(n+sz); }
/* S_N = 2(s+z)/(n+s+z) */

static double sokal_sneath_3  (double s, double r, double n, double *p)
{ double q = r-s; return (q > 0) ? (n-q)/q : INFINITY; }
/* S_O = (s+z)/q = (n-q)/q */

static double baroni_buser    (double s, double r, double n, double *p)
{ double x = sqrt(s*(n-r)); return (x+s)/(x+r); }
/* S_B = (x+s)/(x+r) */

/*--------------------------------------------------------------------*/

static double generic_0       (double s, double r, double n, double *p)
{                               /* --- generic similarity measure */
  double x = sqrt(s *r);
  double d = p[4]*s +p[5]*r +p[6]*n +p[7]*x;
  if (fabs(d) <= DBL_MIN) return INFINITY;
  return (p[0]*s +p[1]*r +p[2]*n +p[3]*x) /d;
}  /* generic_0() */
/* S = (c_0s +c_1z +c_2n +c_3x) / (c_4s +c_5z +c_6n +c_7x) */

/*--------------------------------------------------------------------*/

static double generic_1       (double s, double r, double n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*s +p[5]*r +p[6]*n;
  if (fabs(d) <= DBL_MIN) return INFINITY;
  return (p[0]*s +p[1]*r +p[2]*n) /d;
}  /* generic_1() */
/* S = (c_0s +c_1z +c_2n) / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static double generic_2       (double s, double r, double n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*s +p[5]*r +p[6]*n;
  return (fabs(d) > DBL_MIN) ? (p[0]*s +p[1]*r) /d : INFINITY;
}  /* generic_2() */
/* S = (c_0s +c_1z) / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static double generic_3       (double s, double r, double n, double *p)
{                               /* --- generic similarity measure */
  double d = p[4]*s +p[5]*r +p[6]*n;
  return (fabs(d) > DBL_MIN) ? p[0]*s /d : INFINITY;
}  /* generic_3() */
/* S = c_0s / (c_4s +c_5z +c_6n) */

/*--------------------------------------------------------------------*/

static SIMFN *simfns[] = {
  /* SM_NONE             0 */  constant,
  /* SM_RUSSEL_RAO       1 */  russel_rao,
  /* SM_KULCZYNSKI       2 */  kulczynski,
  /* SM_JACCARD          3 */  jaccard,
  /* SM_DICE             4 */  dice,
  /* SM_SOKAL_SNEATH_1   5 */  sokal_sneath_1,
  /* SM_SOKAL_MICHENER   6 */  sokal_michener,
  /* SM_FAITH            7 */  faith,
  /* SM_ROGERS_TANIMOTO  8 */  rogers_tanimoto,
  /* SM_SOKAL_SNEATH_2   9 */  sokal_sneath_2,
  /* SM_SOKAL_SNEATH_3  10 */  sokal_sneath_3,
  /* SM_BARONI_BUSER    11 */  baroni_buser,
  /* SM_GENERIC         12 */  generic_0,
};

/*----------------------------------------------------------------------
  Auxiliary Functions (for debugging)
----------------------------------------------------------------------*/
#ifndef NDEBUG

static void show_tl (const char *text, TID **lists, ITEM n)
{                               /* --- show transaction database */
  ITEM i;                       /* loop variable */
  TID  *t;                      /* to traverse transaction ids */

  printf("%s\n", text);         /* print the given text */
  for (i = 0; i < n; i++) {     /* traverse items and tid lists */
    printf("%s(%"ITEM_FMT"):", ib_name(ibase, i), i);
    for (t = lists[i]; *t >= 0; t++)
      printf(" %"TID_FMT, *t);  /* print the item name and */
    fputc('\n', stdout);        /* the transaction identifiers */
  }                             /* and terminate the output line */
}  /* show_tl() */

/*--------------------------------------------------------------------*/

static void show_da (const char *text, double *a, size_t n)
{                               /* --- show a double array */
  size_t i;                     /* loop variable */

  printf("%s:", text);          /* print the given text */
  for (i = 0; i < n; i++)       /* traverse the array */
    printf(" %g", a[i]);        /* and print its elements */
  printf("\n");                 /* terminate the output line */
}  /* show_da() */

#endif
/*----------------------------------------------------------------------
  Subset Size Occurrence Distribution Based Item Set Mining
----------------------------------------------------------------------*/

static int ispseudo (RECDATA *rd, double s)
{                               /* --- check a found item set */
  ITEM   n, k, i;               /* loop variables, item buffer */
  ITEM   min;                   /* minimum number of cont. items */
  TID    *t;                    /* to traverse transaction ids */
  double r;                     /* evaluation result */

  assert(rd);                   /* check the function arguments */
  n = isr_cnt(rd->report);      /* get the size of the item set */
  if (n <= 1) return 0;         /* singletons are always valid */
  if (rd->sfrac > 0) {          /* if to use transaction weights */
    s -= (rd->dist[0] -rd->dist[1]) *rd->xsps[n];
    s *= rd->sfrac;             /* get minimum conditional support */
    for (k = n; --k >= 0; ) {   /* traverse the items in the set */
      i = isr_itemx(rd->report, k);
      for (r = 0, t = rd->lists[i]; *t >= 0; t++) {
        r += (double)rd->muls[*t] *rd->xsps[n-rd->cnts[*t]];
        /* if (r > s) break; */ /* traverse the containing trans. */
      }                         /* and check the conditional support */
      if (r < s) return -1;     /* if insufficient, abort function */
    } }
  else {                        /* if to use only trans. counters */
    min = n -rd->xspcnt; if (min <= 0) min = 1;
    s   = -rd->sfrac *rd->dist[min];
    for (k = n; --k >= 0; ) {   /* traverse the items in the set */
      i = isr_itemx(rd->report, k);
      for (r = 0, t = rd->lists[i]; *t >= 0; t++) {
        if (rd->cnts[*t] < min) continue;
        r += (double)rd->muls[*t];
        /* if (r > s) break; */ /* traverse the containing trans. */
      }                         /* and check the conditional support */
      if (r < s) return -1;     /* if insufficient, abort function */
    }
  }
  return 0;                     /* return 'item set is valid' */
}  /* ispseudo() */

/*--------------------------------------------------------------------*/

static int recurse (RECDATA *rd, ITEM n)
{                               /* --- sodim recursion */
  int    r = 0;                 /* error status */
  ITEM   i, k;                  /* loop variables, error status */
  TID    *t;                    /* to traverse the tid lists */
  double s, x;                  /* support of current item set */
  double e;                     /* distribution evaluation */

  assert((n > 0) && rd);        /* check the function arguments */
  k = isr_cnt(rd->report) +1;   /* get the (new) item set size */
  while (--n >= 0) {            /* traverse the items/tid lists */
    for (t = rd->lists[n]; *t >= 0; t++)
      rd->dist[++rd->cnts[*t]] += (double)rd->muls[*t];
    s = x = rd->dist[k];        /* update the support distribution */
    if (s >= rd->smin) {        /* if the item set is frequent */
      if (rd->xsps[0] > 0) {    /* if to use extended support */
        for (i = rd->xspcnt; --i >= 0; )
          x += (double)(rd->dist[k-i-1] -rd->dist[k-i]) *rd->xsps[i+1];
      }                         /* compute the extended support */
      e = rd->simfn(x, rd->dist[1], rd->dist[0], rd->simps);
      if (e >= rd->sim) {       /* if the item set qualifies */
        r = isr_addwgt(rd->report, n, (SUPP)s, e);
        if (r < 0) break;       /* add current item to the reporter */
        if (r > 0) {            /* check if item needs processing */
          if ((n > 0)           /* if the item set can be extended */
          &&  isr_xable(rd->report, 1)
          &&  ((r = recurse(rd, n)) < 0))
            break;              /* find item sets recursively */
          if ((rd->sfrac == 0) || !ispseudo(rd, x)) {
            r = isr_reporto(rd->report, rd->cnts, 0);
            if (r < 0) break;   /* report the current item set */
          }                     /* and check for an error */
          isr_remove(rd->report, 1);
        }                       /* remove the current item */
      }                         /* from the item set reporter */
    }
    for (t = rd->lists[n]; *t >= 0; t++)
      rd->dist[rd->cnts[*t]--] -= (double)rd->muls[*t];
  }                             /* restore the support distribution */
  return r;                     /* return the error status */
}  /* recurse() */

/*--------------------------------------------------------------------*/

int sodim_base (TABAG *tabag, SUPP smin, double *xsps, int xspcnt,
                int simcov, double *simps, double sim, double sfrac,
                ISREPORT *report)
{                               /* --- search for frequent item sets */
  int        r = 0;             /* result of recursion/error status */
  ITEM       i, k;              /* loop variable, number of items */
  TID        n, m;              /* number of transactions */
  size_t     x;                 /* size of the counter array */
  SUPP       w;                 /* weight/support buffer */
  TRACT      *t;                /* to traverse the transactions */
  TID        *tids, *p, **next; /* to traverse transaction ids */
  const ITEM *s;                /* to traverse transaction items */
  const TID  *c;                /* item occurrence counters */
  RECDATA    rd;                /* recursion data */

  assert(tabag && report        /* check the function arguments */
  &&    (xsps || (xspcnt <= 0)) && simps);
  if (simcov < SOD_GENERIC)     /* if specific similarity measure, */
    rd.simfn = simfns[simcov];  /* note the corresponding function */
  else {                        /* if generic similarity measure */
    rd.simps[0] = simps[0];          rd.simps[4] = simps[4];
    rd.simps[1] =         -simps[1]; rd.simps[5] =         -simps[5];
    rd.simps[2] = simps[2]+simps[1]; rd.simps[6] = simps[6]+simps[5];
    rd.simps[3] = simps[3];          rd.simps[7] = simps[7];
    if      ((fabs(rd.simps[3]) > DBL_MIN)
    ||       (fabs(rd.simps[7]) > DBL_MIN))
         rd.simfn = generic_0;  /* use all parameters */
    else if  (fabs(rd.simps[2]) > DBL_MIN)
         rd.simfn = generic_1;  /* do not use geometric mean of s & z */
    else if  (fabs(rd.simps[1]) > DBL_MIN)
         rd.simfn = generic_2;  /* do not use zero zero counter */
    else rd.simfn = generic_3;  /* use normalized support */
  }                             /* (set similarity measure function) */
  rd.smin = smin;               /* note the minimum support */
  isr_setwgt(report, 1);        /* set the base similarity value */
  rd.sim  = (sim > 0) ? sim : DBL_MIN;  /* and the minimum value */
  k = tbg_itemcnt(tabag);       /* get the number of items and */
  if (k <= 0) return 0;         /* check for at least one item */
  c = tbg_icnts(tabag, 0);      /* get the number of containing */
  if (!c) return -1;            /* transactions per item and */
  n = tbg_cnt(tabag);           /* the number of transactions */
  rd.lists = (TID**)malloc((size_t)(k+k) *sizeof(TID*)
                          +(size_t) n    *sizeof(SUPP)
                          +(size_t) n    *sizeof(ITEM));
  if (!rd.lists) return -1;     /* allocate working memory/arrays */
  next    = rd.lists +k;        /* and split off next position array */
  rd.muls = (SUPP*)(next   +k); /* and transaction multiplicity */
  rd.cnts = (ITEM*)(rd.muls+n); /* and item counter array */
  x = (size_t)(((ITEM)xspcnt > k) ? (ITEM)xspcnt : k) +2;
  rd.xsps = (double*)malloc((x+(size_t)(k+2))*sizeof(double));
  if (!rd.xsps) { free(rd.lists); return -1; }
  rd.dist = rd.xsps +x;         /* get the distribution array */
  while (--x > (size_t)xspcnt) rd.xsps[x] = 0;
  for (++x; --x > 0; ) rd.xsps[x] = xsps[x-1];
  rd.xsps[0] = 1.0;             /* copy the subset size weights */
  rd.xspcnt  = xspcnt;          /* and note their number */
  rd.sfrac   = sfrac;           /* copy and fill minimum support, */
  w = tbg_wgt(tabag);           /* note cond. support fraction and */
  for (i = xspcnt+1; --i >= 0;) /* set background support values */
    rd.dist[-i] = (double)w;    /* to avoid index range checks */
  memset(rd.dist+1, 0, (size_t)k *sizeof(double));
  x = tbg_extent(tabag);        /* get the number of item occurrences */
  p = tids = (TID*)malloc(((size_t)k+x) *sizeof(TID));
  if (!p) { free(rd.xsps); free(rd.lists); return -1; }
  for (i = 0; i < k; i++) {     /* traverse the items and */
    rd.lists[i] = next[i] = p;  /* set the transaction id list */
    p += c[i]; *p++ = (TID)-1;  /* skip space for transaction ids */
  }                             /* and store a sentinel at the end */
  for (m = 0; m < n; m++) {     /* traverse the transactions */
    t = tbg_tract(tabag, m);    /* and note their multiplicity, */
    rd.muls[m] = ta_wgt(t);     /* then traverse the items */
    for (s = ta_items(t); *s >= 0; s++)
      *next[*s]++ = m;          /* collect the transaction ids */
  }                             /* in lists per item */
  rd.report = report;           /* note the item set reporter */
  memset(rd.cnts, 0, (size_t)n *sizeof(int));
  r = recurse(&rd, k);          /* find items sets recursively */
  free(tids);                   /* delete the allocated arrays */
  free(rd.xsps); free(rd.lists);
  return r;                     /* return the error status */
}  /* sodim_base() */

/*----------------------------------------------------------------------
  SODIM (generic)
----------------------------------------------------------------------*/

int sod_data (TABAG *tabag, int target, SUPP smin, ITEM zmin,
              int eval, int algo, int mode, int sort)
{                               /* --- prepare data for SODIM */
  ITEM    m;                    /* number of items */
  #ifndef QUIET                 /* if to print messages */
  TID     n;                    /* number of transactions */
  SUPP    w;                    /* total transaction weight */
  clock_t t;                    /* timer for measurements */
  #endif                        /* (only needed for messages) */

  /* --- sort and recode items --- */
  CLOCK(t);                     /* start timer, print log message */
  XMSG(stderr, "filtering, sorting and recoding items ... ");
  m = tbg_recode(tabag, smin, -1, -1, sort);
  if (m <  0) return E_NOMEM;   /* recode items and transactions */
  if (m <= 0) return E_NOITEMS; /* and check the number of items */
  XMSG(stderr, "[%"ITEM_FMT" item(s)]", m);
  XMSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- sort and reduce transactions --- */
  CLOCK(t);                     /* start timer, print log message */
  XMSG(stderr, "sorting and reducing transactions ... ");
  tbg_itsort(tabag, +1, 0);     /* sort items in transactions and */
  if (!(mode & SOD_TIDS)) {     /* if not to report transaction ids, */
    tbg_sort(tabag, +1, 0);     /* sort the trans. lexicographically */
    tbg_reduce(tabag, 0);       /* reduce transactions to unique ones */
  }                             /* (no reduction if tids requested) */
  #ifndef QUIET                 /* if to print messages */
  n = tbg_cnt(tabag);           /* get the number of transactions */
  w = tbg_wgt(tabag);           /* and the total transaction weight */
  XMSG(stderr, "[%"TID_FMT, n); /* print number of transactions */
  if (w != (SUPP)n) { XMSG(stderr, "/%"SUPP_FMT, w); }
  XMSG(stderr, " transaction(s)] done [%.2fs].\n", SEC_SINCE(t));
  #endif
  return 0;                     /* return 'ok' */
}  /* sod_data() */

/*--------------------------------------------------------------------*/

int sod_report (ISREPORT *report, int target,
                int eval, double thresh, int algo, int mode)
{                               /* --- prepare reporter for SODIM */
  assert(report);               /* check the function arguments */
  switch (eval) {               /* if to use an evaluation measure */
    case SOD_LDRATIO:           /* binary log. of support quotient */
      isr_seteval(report, isr_logrto,  NULL, +1, thresh); break;
    case SOD_SIZESIM:           /* item set size times similarity */
      isr_seteval(report, isr_sizewgt, NULL, +1, thresh); break;
  }                             /* set the evaluation function */
  return (isr_settarg(report, target, 0, -1)) ? E_NOMEM : 0;
}  /* sod_report() */

/*--------------------------------------------------------------------*/

int sodim (TABAG *tabag, SUPP smin, double *xsps, int xspcnt,
           int simcov, double *simps, double sim, double sfrac,
           int eval, double thresh, ITEM miscnt,
           int algo, int mode, ISREPORT *report)
{                               /* --- search for frequent item sets */
  int     r;                    /* result of function call */
  #ifndef QUIET                 /* if to print messages */
  clock_t t;                    /* timer for measurements */
  #endif                        /* (only needed for messages) */

  assert(tabag && report);      /* check the function arguments */
  CLOCK(t);                     /* start time, print log message */
  XMSG(stderr, "writing %s ... ", isr_name(report));
  while ((xspcnt > 0) && (xsps[xspcnt-1] <= 0))
    xspcnt--;                   /* remove trailing zeros/negatives */
  if (miscnt < 0)               /* if no accepted number of missing */
    miscnt = (ITEM)xspcnt;      /* items is given, set default */
  isr_tidcfg(report, ib_cnt(tbg_base(tabag)), miscnt);
  r = sodim_base(tabag, smin, xsps, xspcnt,
                 simcov, simps, sim, sfrac, report);
  if (r < 0) return E_NOMEM;    /* find frequent item sets */
  XMSG(stderr, "[%"ISCNT_FMT" set(s)]", ISCOUT(isr_repcnt(report)));
  XMSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  return 0;                     /* return 'ok' */
}  /* sodim() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

static void help (void)
{                               /* --- print add. option information */
  #ifndef QUIET
  fprintf(stderr, "\n");        /* terminate startup message */
  printf("item cover similarity measures (option -c#)\n");
  printf("  r   Russel & Rao                S_R = s/n\n");
  printf("  k   Kulczynski                  S_K = s/q\n");
  printf("  j   Jaccard/Tanimoto            S_J = s/r\n");
  printf("  d   Dice/Sorensen/Czekanowski   S_D = 2s/(r+s)\n");
  printf("  s   Sokal & Sneath 1            S_S = s/(r+q)\n");
  printf("  m   Sokal & Michener/Hamming    S_M = (s+z)/n\n");
  printf("  f   Faith                       S_F = (s+z/2)/n\n");
  printf("  t   Rogers & Tanimoto           S_T = (s+z)/(n+q)\n");
  printf("  n   Sokal & Sneath 2            S_N = 2(s+z)/(n+s+z)\n");
  printf("  o   Sokal & Sneath 3            S_O = (s+z)/q \n");
  printf("  b   Baroni-Urbani & Buser       "
               "S_B = (sqrt(sz)+s)/(sqrt(sz)+r)\n");
  printf("  g   generic similarity measure  "
               "S   = (c0s +c1z +c2n +c3sqrt(sz))\n");
  printf("                                  "
               "    / (c4s +c5z +c6n +c7sqrt(sz))\n");
  printf("      parameters can be specified "
               "with the option -p#:#:#:#:#:#:#:#\n");
  printf("all measures are defined in the following quantities:\n");
  printf("  n   total number of transactions\n");
  printf("  s   number of transactions "
               "with all items in the item set\n");
  printf("  r   number of transactions "
               "with at least one item in the item set\n");
  printf("  q   number of transactions "
               "with some (but not all) items\n");
  printf("  z   number of transactions "
               "with no items in the item set\n");
  printf("\n");
  printf("additional evaluation measures (option -e#)\n");
  printf("  x   no measure (default)\n");
  printf("  b   binary logarithm of support quotient\n");
  printf("  w   similarity measure weighted with itemset size\n");
  printf("\n");
  printf("information output format characters (option -v#)\n");
  printf("  %%%%  a percent sign\n");
  printf("  %%i  number of items (item set size)\n");
  printf("  %%a  absolute item set support\n");
  printf("  %%s  relative item set support as a fraction\n");
  printf("  %%S  relative item set support as a percentage\n");
  printf("  %%w  item cover similarity as a fraction\n");
  printf("  %%W  item cover similarity as a percentage\n");
  printf("  %%z  item cover similarity times number of transactions\n");
  printf("  %%e  additional evaluation measure\n");
  printf("  %%E  additional evaluation measure as a percentage\n");
  printf("  %%Q  total transaction weight (database size)\n");
  printf("All format characters can be preceded by the number\n");
  printf("of significant digits to be printed (at most 32 digits),\n");
  printf("even though this value is ignored for integer numbers.\n");
  #endif                        /* print help information */
  exit(0);                      /* abort the program */
}  /* help() */

/*--------------------------------------------------------------------*/

static ITEM getbdr (char *s, char **end, double **border)
{                               /* --- get the support border */
  ITEM   i, k;                  /* loop variables */
  double *b;                    /* support border */

  assert(s && end && border);   /* check the function arguments */
  for (i = k = 0; s[i]; i++)    /* traverse the string and */
    if (s[i] == ':') k++;       /* count the number separators */
  *border = b = (double*)malloc((size_t)++k *sizeof(double));
  if (!b) return -1;            /* allocate a support border */
  for (i = 0; i < k; i++) {     /* traverse the parameters */
    b[i] = strtod(s, end);      /* get the next parameter and */
    if (*end == s) break;       /* check for an empty parameter */
    s = *end; if (*s++ != ':') break;
  }                             /* check for a colon */
  if (++i < k)                  /* shrink support array if possible */
    *border = (double*)realloc(b, (size_t)i *sizeof(double));
  return i;                     /* return number of support values */
}  /* getbdr() */

/*--------------------------------------------------------------------*/

static int setbdr (ISREPORT *report, SUPP w, ITEM zmin,
                   double **border, ITEM n)
{                               /* --- set the support border */
  double s;                     /* to traverse the support values */

  assert(report                 /* check the function arguments */
  &&    (w > 0) && (zmin >= 0) && border && (*border || (n <= 0)));
  while (--n >= 0) {            /* traverse the support values */
    s = (*border)[n];           /* transform to absolute count */
    s = ceilsupp((s >= 0) ? s/100.0 *(double)w *(1-DBL_EPSILON) : -s);
    if (isr_setbdr(report, n+zmin, (RSUPP)s) < 0) return -1;
  }                             /* set support in item set reporter */
  if (*border) { free(*border); *border = NULL; }
  return 0;                     /* return 'ok' */
}  /* setbdr() */

/*--------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (twrite) twr_delete(twrite, 1); \
  if (report) isr_delete(report, 0); \
  if (tabag)  tbg_delete(tabag,  0); \
  if (tread)  trd_delete(tread,  1); \
  if (ibase)  ib_delete (ibase);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/
#ifdef SODIM_MAIN

int main (int argc, char *argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse the options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_inp  = NULL;      /* name of input  file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *fn_sel  = NULL;      /* name of item selection file */
  CCHAR   *fn_tid  = NULL;      /* name of transaction ids file */
  CCHAR   *fn_psp  = NULL;      /* name of pattern spectrum file */
  CCHAR   *recseps = NULL;      /* record  separators */
  CCHAR   *fldseps = NULL;      /* field   separators */
  CCHAR   *blanks  = NULL;      /* blank   characters */
  CCHAR   *comment = NULL;      /* comment characters */
  CCHAR   *hdr     = "";        /* record header  for output */
  CCHAR   *sep     = " ";       /* item separator for output */
  CCHAR   *dflt    = " (%S)";   /* default format for check */
  CCHAR   *info    = dflt;      /* format for information output */
  int     target   = 's';       /* target type (closed/maximal) */
  ITEM    zmin     = 2;         /* minimum size of an item set */
  ITEM    zmax     = ITEM_MAX;  /* maximum size of an item set */
  double  supp     = 10;        /* minimum support of an item set */
  SUPP    smin     = 1;         /* minimum support of an item set */
  int     simcov   = 'r';       /* item cover similarity measure */
  double  simps[8] = { 1, 0, 0, 0,  0, -1, 1, 0 };
  double  sim      = 10;        /* minimum similarity (in percent) */
  double  *xsps    = NULL;      /* extended support parameters */
  int     xspcnt   = 0;         /* number of ext. support parameters */
  double  sfrac    = 0;         /* minimum cond. support fraction */
  int     eval     = 'x';       /* additional evaluation measure */
  double  thresh   = 10;        /* threshold for evaluation measure */
  int     sort     = 1;         /* flag for item sorting and recoding */
  int     mode     = 0;         /* search mode (e.g. tid output) */
  int     mtar     = 0;         /* mode for transaction reading */
  int     scan     = 0;         /* flag for scanable item output */
  int     bdrcnt   = 0;         /* number of support values in border */
  ITEM    miscnt   = -1;        /* accepted number of missing items */
  int     stats    = 0;         /* flag for item set statistics */
  PATSPEC *psp;                 /* collected pattern spectrum */
  ITEM    m;                    /* number of items */
  TID     n;                    /* number of transactions */
  SUPP    w;                    /* total transaction weight */
  #ifndef QUIET                 /* if not quiet version */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no arguments given */
    printf("usage: %s [options] infile [outfile]\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-t#      target type                              "
                    "(default: %c)\n", target);
    printf("         (s: frequent, c: closed, m: maximal item sets)\n");
    printf("-m#      minimum number of items per item set     "
                    "(default: %"ITEM_FMT")\n", zmin);
    printf("-n#      maximum number of items per item set     "
                    "(default: no limit)\n");
    printf("-s#      minimum support of item sets             "
                    "(default: %g%%)\n", supp);
    printf("         (positive: percentage, "
                     "negative: absolute number)\n");
    printf("-i#      item cover similarity measure to use     "
                    "(default: %c)\n", simcov);
    printf("         (use the default for simple extended support)\n");
    printf("-p#:#..  parameters for generic cover similarity  "
                    "(default: 1:0:0:0:0:-1:1:0)\n");
    printf("-c#      minimum cover similarity of an item set  "
                    "(default: %g%%)\n", sim);
    printf("-x#:#..  subset size weights for extended support "
                    "(default: none)\n");
    printf("-l#      minimum conditional support fraction     "
                    "(default: %g%%)\n", sfrac);
    printf("         (positive: respect, "
                    "negative: ignore transaction weight)\n");
    printf("-L#      minimum conditional support percentage   "
                    "(default: %g%%)\n", sfrac*100.0);
    printf("         (positive: respect, "
                    "negative: ignore transaction weight)\n");
    printf("         (-l# and -L# only interpret "
                    "their argument differently)\n");
    printf("-e#      additional evaluation measure            "
                    "(default: none)\n");
    printf("-d#      threshold for add. evaluation measure    "
                    "(default: %g%%)\n", thresh);
    printf("-q#      sort items w.r.t. their frequency        "
                    "(default: %d)\n", sort);
    printf("         (1: ascending, -1: descending, 0: do not sort,\n"
           "          2: ascending, -2: descending w.r.t. "
                    "transaction size sum)\n");
    printf("-F#:#..  support border for filtering item sets   "
                    "(default: none)\n");
    printf("         (list of minimum support values, "
                    "one per item set size,\n");
    printf("         starting at the minimum size, "
                    "as given with option -m#)\n");
    printf("-R#      read an item selection from a file\n");
    printf("-P#      write pattern spectrum to a file\n");
    printf("-Z       print item set statistics "
                    "(number of item sets per size)\n");
    printf("-g       write output in scanable form "
                    "(quote certain characters)\n");
    printf("-h#      record header  for output                "
                    "(default: \"%s\")\n", hdr);
    printf("-k#      item separator for output                "
                    "(default: \"%s\")\n", sep);
    printf("-v#      output format for item set information   "
                    "(default: \"%s\")\n", info);
    printf("-w       integer transaction weight in last field "
                    "(default: only items)\n");
    printf("-r#      record/transaction separators            "
                    "(default: \"\\n\")\n");
    printf("-f#      field /item        separators            "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                       "
                    "(default: \" \\t\\r\")\n");
    printf("-C#      comment characters                       "
                    "(default: \"#\")\n");
    printf("-T#      file to write transaction identifiers to "
                    "(default: none)\n");
    printf("-M#      accepted number of missing items for tid "
                    "(default: 0)\n");
    printf("-!       print additional option information\n");
    printf("infile   file to read transactions from           "
                    "[required]\n");
    printf("outfile  file to write frequent item sets to      "
                    "[optional]\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */
  #endif  /* #ifndef QUIET */
  /* free option characters: ajouyz [A-Z]\[CFPRTM] */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate switches */
          case '!': help();                             break;
          case 't': target = (*s) ? *s++ : 's';         break;
          case 'm': zmin   = (ITEM)strtol(s, &s, 0);    break;
          case 'n': zmax   = (ITEM)strtol(s, &s, 0);    break;
          case 's': supp   =       strtod(s, &s);       break;
          case 'i': simcov = (*s) ? *s++ : 'j';         break;
          case 'p': getdblvec(s, &s, 8, simps);         break;
          case 'c': sim    =       strtod(s, &s);       break;
          case 'x': if ((xspcnt = getdblvar(s, &s, &xsps)) < 0)
                      error(E_NOMEM); else              break;
          case 'l': sfrac  =       strtod(s, &s);       break;
          case 'L': sfrac  =       strtod(s, &s)/100.0; break;
          case 'e': eval   = (*s) ? *s++ : 0;           break;
          case 'd': thresh =       strtod(s, &s);       break;
          case 'F': bdrcnt = getbdr(s, &s, &border);    break;
          case 'R': optarg = &fn_sel;                   break;
          case 'P': optarg = &fn_psp;                   break;
          case 'Z': stats  = 1;                         break;
          case 'g': scan   = 1;                         break;
          case 'h': optarg = &hdr;                      break;
          case 'k': optarg = &sep;                      break;
          case 'v': optarg = &info;                     break;
          case 'q': sort   = (int) strtol(s, &s, 0);    break;
          case 'w': mtar  |= TA_WEIGHT;                 break;
          case 'r': optarg = &recseps;                  break;
          case 'f': optarg = &fldseps;                  break;
          case 'b': optarg = &blanks;                   break;
          case 'C': optarg = &comment;                  break;
          case 'T': optarg = &fn_tid;                   break;
          case 'M': miscnt = (ITEM)strtol(s, &s, 0);    break;
          default : error(E_OPTION, *--s);              break;
        }                       /* set option variables */
        if (optarg && *s) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* -- if argument is no option */
      switch (k++) {            /* evaluate non-options */
        case  0: fn_inp = s;      break;
        case  1: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg)       error(E_OPTARG);     /* check option arguments */
  if (k      < 1)   error(E_ARGCNT);     /* and number of arguments */
  if (zmin   < 1)   error(E_SIZE, zmin); /* check the size limits */
  if (zmax   < 0)   error(E_SIZE, zmax); /* and the minimum support */
  if (supp   > 100) error(E_SUPPORT, supp);
  if (bdrcnt < 0)   error(E_NOMEM);
  if ((!fn_inp || !*fn_inp) && (fn_sel && !*fn_sel))
    error(E_STDIN);             /* stdin must not be used twice */
  switch (target) {             /* check and translate target type */
    case 's': target = ISR_ALL;               break;
    case 'c': target = ISR_CLOSED;            break;
    case 'm': target = ISR_MAXIMAL;           break;
    default : error(E_TARGET, (char)target);  break;
  }                             /* (get target type code) */
  switch (simcov) {             /* check and translate measure */
    case 'r': simcov = SOD_RUSSEL_RAO;        break;
    case 'k': simcov = SOD_KULCZYNSKI;        break;
    case 'j': simcov = SOD_JACCARD;           break;
    case 'd': simcov = SOD_DICE;              break;
    case 's': simcov = SOD_SOKAL_SNEATH_1;    break;
    case 'm': simcov = SOD_SOKAL_MICHENER;    break;
    case 'f': simcov = SOD_FAITH;             break;
    case 't': simcov = SOD_ROGERS_TANIMOTO;   break;
    case 'n': simcov = SOD_SOKAL_SNEATH_2;    break;
    case 'o': simcov = SOD_SOKAL_SNEATH_3;    break;
    case 'b': simcov = SOD_BARONI_BUSER;      break;
    case 'g': simcov = SOD_GENERIC;           break;
    default : error(E_SIMILTY, (char)simcov); break;
  }                             /* (get cover similarity measure) */
  switch (eval) {               /* check and translate measure */
    case 'x': eval = SOD_NONE;               break;
    case 'b': eval = SOD_LDRATIO;            break;
    case 'w': eval = SOD_SIZESIM;            break;
    default : error(E_MEASURE, (char)eval);  break;
  }                             /* (get evaluation measure code) */
  if (fn_tid) {                 /* if to write transaction ids. */
    if (strcmp(fn_tid, "-") == 0) fn_tid = "";
    mode |= SOD_TIDS;           /* turn "-" into "" for consistency */
  }                             /* set transaction identifier flag */
  if (info == dflt)             /* adapt the default info. format */
    info = (supp < 0) ? " (%a)" : " (%S)";
  thresh /= 100.0;              /* scale evaluation threshold */
  MSG(stderr, "\n");            /* terminate the startup message */

  /* --- read item selection --- */
  ibase = ib_create(0, 0);      /* create an item base */
  if (!ibase) error(E_NOMEM);   /* to manage the items */
  tread = trd_create();         /* create a transaction reader */
  if (!tread) error(E_NOMEM);   /* and configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_sel) {                 /* if item appearances are given */
    CLOCK(t);                   /* start timer, open input file */
    if (trd_open(tread, NULL, fn_sel) != 0)
      error(E_FOPEN, trd_name(tread));
    MSG(stderr, "reading %s ... ", trd_name(tread));
    m = ib_readsel(ibase,tread);/* read the given item selection */
    if (m < 0) error((int)-m, ib_errmsg(ibase, NULL, 0));
    trd_close(tread);           /* close the input file */
    MSG(stderr, "[%"ITEM_FMT" item(s)]", m);
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* print a log message */

  /* --- read transaction database --- */
  tabag = tbg_create(ibase);    /* create a transaction bag */
  if (!tabag) error(E_NOMEM);   /* to store the transactions */
  CLOCK(t);                     /* start timer, open input file */
  if (trd_open(tread, NULL, fn_inp) != 0)
    error(E_FOPEN, trd_name(tread));
  MSG(stderr, "reading %s ... ", trd_name(tread));
  k = tbg_read(tabag, tread, mtar);
  if (k < 0) error(-k, tbg_errmsg(tabag, NULL, 0));
  trd_delete(tread, 1);         /* read the transaction database, */
  tread = NULL;                 /* then delete the table reader */
  m = ib_cnt(ibase);            /* get the number of items, */
  n = tbg_cnt(tabag);           /* the number of transactions, */
  w = tbg_wgt(tabag);           /* the total transaction weight */
  MSG(stderr, "[%"ITEM_FMT" item(s), %"TID_FMT, m, n);
  if (w != (SUPP)n) { MSG(stderr, "/%"SUPP_FMT, w); }
  MSG(stderr, " transaction(s)] done [%.2fs].", SEC_SINCE(t));
  if ((m <= 0) || (n <= 0))     /* check for at least one item */
    error(E_NOITEMS);           /* and at least one transaction */
  MSG(stderr, "\n");            /* terminate the log message */
  supp = (supp >= 0) ? supp/100.0 *(double)w *(1-DBL_EPSILON) : -supp;
  smin = (SUPP)ceilsupp(supp);  /* compute absolute support value */
  sim  = (sim >= 0) ? sim/100.0 : -sim/(double)w;

  /* --- find frequent item sets --- */
  mode |= SOD_VERBOSE|SOD_NOCLEAN;
  k = sod_data(tabag, target, smin, zmin, eval, SOD_BASIC, mode, sort);
  if (k) error(k);              /* prepare data for SODIM */
  report = isr_create(ibase);   /* create an item set reporter */
  if (!report) error(E_NOMEM);  /* and configure it */
  isr_setsize(report, zmin, zmax);
  isr_setsupp(report, smin, SUPP_MAX);
  if (setbdr(report, w, zmin, &border, bdrcnt) != 0)
    error(E_NOMEM);             /* set limits and support border */
  if (fn_psp && (isr_addpsp(report, NULL) < 0))
    error(E_NOMEM);             /* set a pattern spectrum if req. */
  if (isr_setfmt(report, scan, hdr, sep, NULL, info) != 0)
    error(E_NOMEM);             /* set the output format strings */
  k = isr_open(report, NULL, fn_out);
  if (k) error(k, isr_name(report));     /* open item set file */
  if (isr_tidopen(report, NULL, fn_tid) != 0)
    error(E_FOPEN, isr_tidname(report)); /* open trans. id file */
  if ((sod_report(report, target, eval, thresh, SOD_BASIC, mode) < 0)
  ||  (isr_setup(report) < 0))  /* prepare reporter for Eclat and */
    error(E_NOMEM);             /* set up the item set reporter */
  k = sodim(tabag, smin, xsps, xspcnt,
            simcov, simps, sim, sfrac, eval, thresh,
            miscnt, SOD_BASIC, mode, report);
  if (stats)                    /* print item set statistics */
    isr_prstats(report, stdout, 0);
  if (isr_close   (report) != 0)/* close item set output file */
    error(E_FWRITE, isr_name(report));
  if (isr_tidclose(report) != 0)/* close trans. id output file */
    error(E_FWRITE, isr_tidname(report));

  /* --- write pattern spectrum --- */
  if (fn_psp) {                 /* if to write a pattern spectrum */
    CLOCK(t);                   /* start timer, create table write */
    psp    = isr_getpsp(report);/* get the pattern spectrum */
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* open the output file */
    if (twr_open(twrite, NULL, fn_psp) != 0)
      error(E_FOPEN,  twr_name(twrite));
    MSG(stderr, "writing %s ... ", twr_name(twrite));
    if (psp_report(psp, twrite, 1.0) != 0)
      error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* write the pattern spectrum */
    twrite = NULL;              /* and delete the table writer */
    MSG(stderr, "[%"SIZE_FMT" signature(s)]", psp_sigcnt(psp));
    MSG(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                             /* write a log message */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */

#endif
